#include "CtfData.h"
#include "fit.h"
#include <qmsgbox.h>
#include <qfiledialog.h> 
#include <math.h>
#include "CtfDatab.C"

CtfData::CtfData(int ctst) {
int i;

//printf("%d\n",ctst);
if (ctst>1000||ctst<10) ctst=CTFDATASTEPS;

CTFSTEPS= ctst;

Lname="New";

file="/";

defocus=0;
B=200.0;
amp=1.0;
ampcont=0.1;
exps4=0;
noise1=0;
noise2=1.0;
noise3=0;
noise4=1.0;
stigang=0;
stigdf=0;
cfilmsf=0;
structype=0;
V=300.0;
Cs=1.6;
sstep=2.333;
smax=1.0/8.0;
ctfmap=NULL;
nptcl=0;
dose=0;
mag=50.;
bgmode=0;

scanner="Default";
microscope="Default";

Flags=CDT_HASEQ|CDT_PLOT;

data=new DataSet;
realimage=fftimage=NULL;
cmgamma=new Equation;
cmctf=new Equation;

data->setRC(16,CTFSTEPS);
// s
data->getColumn(0)->setRange(smax/(float)CTFSTEPS,smax,CTFSTEPS);

// gamma
data->getColumn(1)->setEqn("-2*pi*(Cs*1.0e7/4*lambda^3*x0^4+defocus*5000*lambda*x0^2)",
	CTFSTEPS,data);
cmgamma->parseEqn("-2*pi*(Cs*1.0e7/4*lambda^3*S^4+defocus*5000*lambda*S^2)");

// gamma, stigmatic
data->getColumn(2)->setEqn("-2*pi*(Cs*1.0e7/4*lambda^3*x0^4+(defocus+stigdf)*5000*lambda*x0^2)",
	CTFSTEPS,data);

// CTF (not squared)
data->getColumn(3)->setEqn(
	"amp*exp(-(exps4*B^.5*x0+B*(1-exps4)*x0^2))*(sqrt(1-ampcont^2)*sin(x1)+ampcont*cos(x1))",
	CTFSTEPS,data);
cmctf->parseEqn("amp*exp(-(exps4*B^.5*S+B*(1-exps4)*S^2))*(sqrt(1-ampcont^2)*sin(gamma)+ampcont*cos(gamma))");

/*data->getColumn(3)->setEqn(
	"amp*exp(-B*(exps4*B*x0^4+(1-exps4)*x0^2))*(sqrt(1-ampcont^2)*sin(x1)+ampcont*cos(x1))",
	CTFSTEPS,data);*/

// CTF, stigmatic
data->getColumn(4)->setEqn(
	"amp*exp(-(exps4*B^.5*x0+B*(1-exps4)*x0^2))*(sqrt(1-ampcont^2)*sin(x2)+ampcont*cos(x2))",
	CTFSTEPS,data);

/*data->getColumn(4)->setEqn(
	"amp*exp(-B*(exps4*B*x0^4+(1-exps4)*x0^2))*(sqrt(1-ampcont^2)*sin(x1)+ampcont*cos(x1))",
	CTFSTEPS,data);*/

// noise (intensity)
data->getColumn(5)->setEqn(
	"(noise3*exp(-((pi/2*noise4*x0)^2)-x0*noise2-sqrt(fabs(x0))*noise1))",CTFSTEPS,
	data);
/* data->getColumn(5)->setEqn(
	"(noise3*exp(-((pi/2*noise4*x0)^2))+noise1*exp(-x0*noise2))^2",CTFSTEPS,
	data); */

/*data->getColumn(5)->setEqn(
	"(noise3*exp(-((pi/2*noise4*x0)^2))+noise1*exp(-x0*noise2))",CTFSTEPS,
	data);*/



// overall CTF
data->getColumn(6)->setEqn("(x13+cfilmsf*x15)*x3^2+x5",CTFSTEPS,data);

// stigmatic overall CTF
data->getColumn(7)->setEqn("(x13+cfilmsf*x15)*x4^2+x5",CTFSTEPS,data);

// SNR - theoretical, calculated from fit parameters and structure factor
data->getColumn(8)->setEqn("x13*x3^2/x5",CTFSTEPS,data);
//data->getColumn(8)->setEqn("x10/x5-1",CTFSTEPS,data);

// Ptcl reqd vs resolution
//data->getColumn(9)->setEqn("2*x0*cfilmsf/x8",CTFSTEPS,data);
data->getColumn(9)->setEqn("2*x0*x5/(amp^2*exp(-(2*B*x0^2)))",CTFSTEPS,data);

// Actual data
data->getColumn(10)->setData((float *)calloc(1,CTFSTEPS*sizeof(float)),CTFSTEPS);

// calculated structure factor
//data->getColumn(11)->setEqn("sqrt(exp(-B*S^2)*(x10-x5)/x3^2)",CTFSTEPS,data);
data->getColumn(11)->setEqn("sqrt((x10-x5)/x3^2)",CTFSTEPS,data);

// stigmatic axis data
data->getColumn(12)->setData((float *)calloc(1,CTFSTEPS*sizeof(float)),CTFSTEPS);

// structure factor from file
data->getColumn(13)->setRange(1.0,1.0,CTFSTEPS);

// User function
data->getColumn(14)->setEqn("x5",CTFSTEPS,data);		// default is noise without signal

data->getColumn(15)->setEqn("less(.01,x0)*10^(3.5037 - 170.42*x0+4034.4*x0^2-52783*x0^3 + 4.1868e+05*x0^4 -2.1402e+06*x0^5 + 7.2256e+06*x0^6 - 1.6012e+07*x0^7 + 2.2365e+07*x0^8 -1.7822e+07*x0^9 + 6.1647e+06*x0^10)",CTFSTEPS,data);

updateEqn();

for (i=0; i<6; i++) {
	plot[i]=new LSPlot;
	plot[i]->setData(data);
	plot[i]->setCols(0,6,1,1);
}

plot[0]->setLabel("New");
}

CtfData::~CtfData() {
int i;

if (realimage) delete realimage;
if (fftimage) delete fftimage;
if (ctfmap) delete ctfmap;

delete data;
for (i=0; i<4; i++) delete plot[i];
}

//The next 3 methods are from ListBoxItem
void CtfData::paint( QPainter *p ) {
	QFontMetrics fm = p->fontMetrics();
	int yPos;                       // vertical text position

	yPos = fm.ascent() + fm.leading()/2;
	p->drawText(16, yPos, lName() );

	if (Flags&CDT_PLOT) {
		p->setPen(QPen(plot[0]->getColor(),3));
		p->drawLine(2,fm.lineSpacing()/2,14,fm.lineSpacing()/2);
	}
}

void CtfData::setBGMode(int mode) {
bgmode=mode;

if (bgmode) {
	data->getColumn(5)->setData((float *)calloc(1,CTFSTEPS*sizeof(float)),CTFSTEPS);
	updateEqn();
}
else {
	data->getColumn(5)->setEqn(
		"(noise3*exp(-((pi/2*noise4*x0)^2)-x0*noise2-sqrt(fabs(x0))*noise1))",CTFSTEPS,
		data);
	updateEqn();
}

}

int CtfData::height(const QListBox *lb ) const {
	return lb->fontMetrics().lineSpacing() + 1 ;
}

int CtfData::width(const QListBox *lb ) const {
	return lb->fontMetrics().width( text() ) + 20;
}


void CtfData::setAEq(int n,QString eq) {		// changes auxilliary equations from default

if (n==-1) cmgamma->parseEqn(eq);
else if (n==-2) cmctf->parseEqn(eq);

}

QString CtfData::getAEq(int n) {
if (n==-1) return cmgamma->getEqn();
else if (n==-2) return cmctf->getEqn();

return QString(NULL);
}

void CtfData::setLName(const QString &s) { 
if (s.findRev('/')>=0 && s.findRev('/',s.findRev('/')-1)>=0) {
	Lname=s.mid(s.findRev('/',s.findRev('/')-1)+1);
}
else Lname=s;

data->title=s;  //commented out by prb

if (s.findRev('/')>=0) plot[0]->setLabel(s.mid(s.findRev('/')+1));
else plot[0]->setLabel(s);
}


void CtfData::setImParm(float V0,float Cs0,float sstep0,float smax0) {
int ui=0;

if (V0!=V||sstep0!=sstep||smax0!=smax) ui=1;

if (V0>1.0&&V0<2000.0) V=V0;
Cs=Cs0;
if (sstep0<1000.0 && sstep0>0) sstep=sstep0;
if (smax0>0&&smax0<10.0) smax=smax0;
if (ui) {
	updateImage();
	if (structype==-1) setStrucMode(structype,strucfile);
}

}

void CtfData::setSigParm(float defocus0,float envwid0, float amp0, float ampcont0, float exps40) {
defocus=defocus0;
B=envwid0;
amp=amp0;
ampcont=ampcont0;
exps4=exps40;

if (B<0) { B*=-1.0; exps4=1.0; }

}

void CtfData::setNoiseParm(float noise10,float noise20,float noise30,float noise40) {
noise1=noise10; 
noise2=noise20; 
noise3=noise30; 
noise4=noise40;
}

void CtfData::setSZero(float r,float r2,float ang) {
float s,s2;

if (!fftimage) return;
s=r/(sstep*fftimage->ySize());
s2=r2/(sstep*fftimage->ySize());
//defocus=(asin(ampcont)/(PI*2.0)-.5-Cs*.25*pow(lambda,3.0)*pow(s,4.0))/(5000.0*lambda*s*s);
if (s>0) defocus=-(2.0*PI+pow(lambda,3.0F)*PI*pow(s,4.0F)*Cs*1.0e7-2.0*asin(ampcont))/(20000.0*lambda*PI*s*s);
if (s2>0) stigdf=-defocus-(2.0*PI+pow(lambda,3.0F)*PI*pow(s2,4.0F)*Cs*1.0e7-2.0*asin(ampcont))/(20000.0*lambda*PI*s2*s2);
stigang=ang;
updateEqn();
}

float CtfData::getZeroR() { 
	float s;
	if (!fftimage) return(0);
	s=1.0/(sstep*fftimage->ySize());
	return zero/s;
}

float CtfData::getZeroStigR() { 
	float s;
	if (!fftimage) return(0);
	s=1.0/(sstep*fftimage->ySize());
	return zerostig/s;
}

void CtfData::setParmStr(const char *s) {
sscanf(s+2,"%f %f %f %f %f %f %f %f %f %f %f",
	&defocus,&B,&amp,&ampcont,&noise1,&noise2,&noise3,&noise4,&V,&Cs,&sstep);

	if (B<0) { B*=-1.0; exps4=1.0; }
smax=.5/sstep;
/*printf("parmstr %f %f %f %f %f %f %f %f %f %f %f",
	defocus,B,amp,ampcont,noise1,noise2,noise3,noise4,V,Cs,sstep);*/
updateEqn();
updateImage();
}

void CtfData::setParm(const float *f) {
defocus=f[0];
B=f[1];
amp=f[2];
ampcont=f[3];
noise1=f[4];
noise2=f[5];
noise3=f[6];
noise4=f[7];
V=f[8];
Cs=f[9];
sstep=f[10];

if (B<0) { B*=-1.0; exps4=1.0; }

smax=.5/sstep;
/*printf("parmstr %f %f %f %f %f %f %f %f %f %f %f",
	defocus,B,amp,ampcont,noise1,noise2,noise3,noise4,V,Cs,sstep);*/
updateEqn();
updateImage();
}

void CtfData::setStigParm(float stigang0, float stigdf0) {
int i=0;

if (stigang!=stigang0 && stigdf0!=0) i=1;
if ((stigdf0>=-.01 && stigdf<-.01) || (stigdf>=-.01 && stigdf0<-.01)) i=1;
stigang=stigang0;
stigdf=stigdf0;
if (i) updateImage();
}

void CtfData::setStrucMode(int mode,QString name) {
structype=mode;
strucfile=name;

if (structype==-1) {
	FILE *in = fopen(strucfile.data(),"rb");
	if (in) {
		float s,*d,fs0,fd0,fs1=-1.0,fd1;
		int i,j;
		
		d=(float *)malloc(CTFSTEPS*sizeof(float));
	
		for (i=0; i<CTFSTEPS; i++) {
			s=data->dataAt(0,i);
			// read until we have data bracketing the correct s value
			while (fs1<s) {
				fs0=fs1; 
				fd0=fd1; 
				if (fscanf(in," %f %f",&fs1,&fd1)!=2) { fs1=-1.0; break; }
			}
			// an exact match
			if (fs1==s) d[i]=fd1;
			else if (fs0==s) d[i]=fd0;
				// no data in the file in the right range
			else if (fs0<0 || fs1<0) { 
				d[i]=1.0; 
				printf("WARNING: structure factor undefined at s=%1.4f\n",s); 
			}
			// linear interpolation
			else {
				d[i]=(fd1*(s-fs0) +fd0*(fs1-s))/(fs1-fs0);
//				printf("%f\t%f\n",s,d[i]);
			}
		}
		data->getColumn(13)->setData(d,CTFSTEPS);
		
	}
	else printf("Warning, SF read failed\n");

}

//if (structype==1) data->getColumn(13)->setEqn("10^-6*(sin(x0*cfilmsf)/x0^3 - cfilmsf*cos(x0*cfilmsf)/x0^2)",CTFSTEPS,data);
if (structype==0) data->getColumn(13)->setRange(1.0,1.0,CTFSTEPS);

}

void CtfData::setStrucParm(float cfilmsf0) {
cfilmsf=cfilmsf0;

if (cfilmsf==0) {

} else {

}

}

void CtfData::setRealImage(EMData *image) {
realimage=image;
if (image) Flags|=CDT_HASREAL; else Flags&=~CDT_HASREAL;
}

void CtfData::setQ(float q1,float q2,float q3,float q4,float q5,float q6) {
Q[0]=q1;
Q[1]=q2;
Q[2]=q3;
Q[3]=q4;
Q[4]=q5;
Q[5]=q6;
}

void CtfData::setFFTImage(EMData *image) {
fftimage=image;
if (image) Flags|=CDT_HASFFT; else Flags&=~CDT_HASFFT;
updateImage();
}

#define WFSIZE	1024
void CtfData::WFilterMRC(const char *fsp ){
EMData ftf,tmp,*cl,*fcl;
float ctfp[11] = { defocus,B,amp,ampcont,noise1,noise2,noise3,noise4,V,Cs,sstep };
float *ctf;
int x,y,i,j;
XYData SF;

if (tmp.readMRC(fsp,1)) error(ERR_EXIT,"Failed to read MRC file");
ftf.setSize(tmp.xSize(),tmp.ySize(),1);
int nx = ftf.xSize();
int ny = ftf.ySize();

if (SF.readFile((const char *)strucfile)) error(ERR_EXIT,"Failed to read SF file");
float *d0=SF.getData();
for (j=1; j<SF.n()*2; j+=2) d0[j]=(d0[j]<=0)?-4.0:log10(d0[j]);
SF.update();

for (y=0; y<ny; y+=WFSIZE) {
	for (x=0; x<nx; x+=WFSIZE) {
		if (x+WFSIZE>nx) x=nx-WFSIZE;
		if (y+WFSIZE>ny) y=ny-WFSIZE;
//		fcl=ftf.clip(x,y,WFSIZE,WFSIZE);
		EMData tf;
		tf.readMRCArea(fsp,x,y,WFSIZE,WFSIZE);
		tf.edgeNormalize();
		cl=tf.clip(-128,-128,WFSIZE+256,WFSIZE+256);

/*char s[256];
sprintf(s,"b4.%d.%d.mrc",x,y);
cl->writeMRC(s);*/

		// high-pass filter		
		cl->setCTF(ctfp);
		ctf=cl->ctfCurve(7,&SF);
//		for (i=0; i<cl->xSize()/2*CTFOS; i++) ctf[i]*=(1.0-exp(-i*i/(SQR(21.0*CTFOS))));
		for (i=0; i<cl->xSize()/2*CTFOS; i++) ctf[i]*=(1.0-exp(-i*i/(SQR(cl->xSize()/40*CTFOS))));
		
		// apply the filter
		fcl=cl->doFFT();
		cl->gimmeFFT();
		fcl->applyRadFn(CTFOS*cl->ySize()/2,0,1.0/CTFOS,ctf,1);
		delete cl;
		cl=fcl->doIFT();
		fcl->gimmeFFT();
		delete fcl;

/*sprintf(s,"aft.%d.%d.mrc",x,y);
cl->writeMRC(s);*/

		// reinsert the filtered region
		fcl=cl->clip(128,128,WFSIZE,WFSIZE);
		ftf.insertClip(fcl,x,y);
		
		delete fcl;
		delete cl;
		free(ctf);
		printf("%d,%d\n",x,y);
	}
}

char *s = (char *) malloc(strlen(fsp)+6);
strcpy(s,fsp);
sprintf(s+strlen(s)-4,".filt.mrc");
ftf.writeMRC(s);
free(s);
}

/*
for (y=0; y<ny-256; y+=WFSIZE) {
	for (x=0; x<nx-256; x+=WFSIZE) {
		if (x+WFSIZE+256>nx) x=nx-WFSIZE-256;
		if (y+WFSIZE+256>ny) y=ny-WFSIZE-256;
//		cl=ftf.clip(x,y,WFSIZE+256,WFSIZE+256);
		tf.readMRCArea(fsp,x,y,WFSIZE+256,WFSIZE+256);
		cl=tf.copy();
		
		cl->setCTF(ctfp);
		ctf=cl->ctfCurve(6,&SF);
		for (i=0; i<12; i++) ctf[i]=0;	// high-pass filter
		ctf[12]*=.333;
		ctf[13]*=.667;
		
		fcl=cl->doFFT();
		cl->gimmeFFT();
		fcl->applyRadFn(CTFOS*cl->ySize(),0,1.0/CTFOS,ctf,1);
		delete cl;
		cl=fcl->doIFT();
		fcl->gimmeFFT();
		delete fcl;
		fcl=cl->clip(128,128,WFSIZE,WFSIZE);
		ftf.insertClip(fcl,x+128,y+128);
		
		delete fcl;
		delete cl;
		free(ctf);
		printf("%d,%d\n",x,y);
	}
}*/


void CtfData::WFilter(){
float *fix;
float s0,ds, SNR,noiseInt,lessthan1Factor;
int i,ny,ny2;
EMData *r,*f,*r2,*rs;
QString nm;
char n[80];

if (!(Flags&CDT_HASFFT)) { warn("This image doesn't have a FFT."); return; }


// Find index corresponding to the first zero


float deltaf = defocus*(1.e-6);
float epsilon = asin((ampcont));
float lam = lambda*1.e-10;
float C_s = Cs*1.e-3;

float FirstZero =  sqrt(lam)* sqrt(       \
	(sqrt( deltaf*deltaf - 2*lam *C_s* ( 1 - epsilon) )  \
	- deltaf )/(2 *( 1 - epsilon) )); 
float spacing = (data->dataAt(0,2)) - (data->dataAt(0,1));
int IFirstZero = 1./(FirstZero*1.e10*spacing);
float FirstZeroAmp = data->dataAt(13,IFirstZero);
 printf("deltaf =%2.2e, FirstZero (Ang)=%2.2e,  spacing=%2.2e, IfirstZero=%d , FirstZeroAmp=%2.2e\n",
 		deltaf,FirstZero*1.e10, spacing,IFirstZero,FirstZeroAmp);

fix=(float *)malloc(CTFSTEPS*sizeof(float));

for (i=0; i<CTFSTEPS; i++) {
//                       %%    NEED TO CHANGE THIS PART (fix)
// if (data->dataAt(3,i)<0) fix[i]=-1.0; else fix[i]=1.0; 
 //   if (data->dataAt(10,i)>0) {
		SNR = (data->dataAt(3,i)) * (data ->dataAt(13,i));
		noiseInt = data->dataAt(5,i);
		SNR= SNR*SNR/noiseInt;
		lessthan1Factor = 2* sqrt(SNR)/(1+SNR);	
		fix[i] =  (data ->dataAt(13,i)) * lessthan1Factor / (2*sqrt(noiseInt)) ;
		if ( data->dataAt(3,i) <0) fix[i] *= -1;
//	} else fix[i]=0; 
//	if (i < IFirstZero) fix[i] *= FirstZeroAmp /(data->dataAt(13,i) )      ;
// printf("i=%d fix[i]=%f fac<1 =%2.2e SNR=%2.2e , SF=%2.2e, 1/noiseInt=%2.2e, 1/m =%2.2e \n",
//    i,fix[i],lessthan1Factor,SNR,data -> dataAt(13,i) , 1/noiseInt, 1/(data->dataAt(3,i)));
 }


s0=data->dataAt(0,0)*sstep*fftimage->ySize();
ds=(data->dataAt(0,1)-data->dataAt(0,0))*sstep*fftimage->ySize();

save_data(s0,ds,fix,CTFSTEPS,"phases");

r=new EMData;
if (file.findRev('.')>0) nm=file.left(file.findRev('.'));
else nm=file;
nm+=".WF";

sprintf(n,"!-%1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g",
	defocus,exps4>.5?-B:B,amp,ampcont,noise1,noise2,noise3,noise4,V,Cs,sstep);
n[79]=0;

updateEqn(11);	// generates a 'sign map' for the current simulation
ny=ctfmap->ySize();

r->readImage(getFile(),0);
ny2=r->ySize();
printf("ny =%d  ny2=%d \n",ny ,ny2);
if (ny!=ny2) { warn("Do NOT use oversampling when performing CTF correction."); return; }

for (i=0; r->readImage(getFile(),i)==0; i++) {
	ny2=r->ySize();
	if (ny!=ny2) rs=r->clip(-(ny-ny2)/2,-(ny-ny2)/2,ny,ny);
	else rs=r->copy();
	rs->normalize();
	f=rs->doFFT();
	rs->gimmeFFT();
	delete rs;
	f->mult(ctfmap);
	r2=f->doIFT();
	f->gimmeFFT();
	r2->clearCTF();
	r2->setName(n);
	if (ny!=ny2) {
		rs=r2->clip((ny-ny2)/2,(ny-ny2)/2,ny2,ny2);
		rs->writeIMAGIC(nm,i);
		delete rs;
	}
	else r2->writeIMAGIC(nm,i);
	delete f;
	delete r2;
	printf("%d done\r",i);
}

free(fix); 
}


void CtfData::fixPhase() {
float *fix;
float s0,ds;
int i,ny,ny2;
EMData *r,*f,*r2,*rs;
QString nm;
char n[80];

if (!(Flags&CDT_HASFFT)) { warn("This image doesn't have a FFT."); return; }


fix=(float *)malloc(CTFSTEPS*sizeof(float));
for (i=0; i<CTFSTEPS; i++) { if (data->dataAt(3,i)<0) fix[i]=-1.0; else fix[i]=1.0; }

s0=data->dataAt(0,0)*sstep*fftimage->ySize();
ds=(data->dataAt(0,1)-data->dataAt(0,0))*sstep*fftimage->ySize();

save_data(s0,ds,fix,CTFSTEPS,"phases");

r=new EMData;
if (file.findRev('.')>0) nm=file.left(file.findRev('.'));
else nm=file;
nm+=".fix";

sprintf(n,"!-%1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g",
	defocus,exps4>.5?-B:B,amp,ampcont,noise1,noise2,noise3,noise4,V,Cs,sstep);
n[79]=0;

updateEqn(1);	// generates a 'sign map' for the current simulation
ny=ctfmap->ySize();

r->readImage(getFile(),0);
ny2=r->ySize();
if (ny!=ny2) { warn("Do NOT use oversampling when performing CTF correction."); return; }

for (i=0; r->readImage(getFile(),i)==0; i++) {
	ny2=r->ySize();
	if (ny!=ny2) rs=r->clip(-(ny-ny2)/2,-(ny-ny2)/2,ny,ny);
	else rs=r->copy();
	rs->normalize();
	f=rs->doFFT();
	rs->gimmeFFT();
	delete rs;
	f->mult(ctfmap);
	r2=f->doIFT();
	f->gimmeFFT();
	r2->clearCTF();
	r2->setName(n);
	if (ny!=ny2) {
		rs=r2->clip((ny-ny2)/2,(ny-ny2)/2,ny2,ny2);
		rs->writeIMAGIC(nm,i);
		delete rs;
	}
	else r2->writeIMAGIC(nm,i);
	delete f;
	delete r2;
	printf("%d done\n",i);
}

free(fix); 
}

void CtfData::setParm() {
int i,i0=0,i1=1;
EMData *r;
char n[80];

if (!(Flags&CDT_HASFFT)) { warn("This image doesn't have a FFT."); return; }

r=new EMData;
sprintf(n,"!-%1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g",
	defocus,exps4>.5?-B:B,amp,ampcont,noise1,noise2,noise3,noise4,V,Cs,sstep);
n[79]=0;
float* ctfarray=getParmArray();
			
if ((i=text().find('('))>0) {
	QString s=text().mid(i+1);
	sscanf((const char *)s,"%d-%d",&i0,&i1);
	i1++;
}
else i1=fileCount(getFile());

for (i=i0; i<i1 && r->readImage(getFile(),i,1)==0; i++) {
	r->clearCTF();
	//r->setName(n);
	r->setCTF(ctfarray);
	//r->writeIMAGIC(file,i);
	r->writeImage(getFile(),i,1);
}

printf("Set parameters on images %d-%d\n",i0,i);

delete r;
delete ctfarray;
}

void CtfData::setSF(const char *name) {
EMData *d;

updateEqn(5,fftimage->ySize());
fftimage->mult(ctfmap);
d=fftimage->doIFT();
QString s(name);
d->writeMRC(name);
delete d;
}

// forward = 1 applies ctf
// forward = -1 does simple ctf correction
// forward = 2 applies ctf with noise (3d)
void CtfData::applyCtfTo(const char *name,int forward) {
EMData *d,*df;
int i,n;
float s0,ds,*noise;

d=new EMData;
d->readImage(name,-1,1);
if (d->zSize()>1) {
	delete d;		// just to make sure no prob with nodata flag
	d=new EMData;
	d->readImage(name,-1);
	df=d->doFFT();
	d->gimmeFFT();
	delete d;
	if (forward==-1) updateEqn(4,d->ySize());
	else updateEqn(3,d->ySize());
	df->mult(ctfmap);
	if (forward==2) {
		noise=data->getColumn(5)->getData();

		s0=data->dataAt(0,0)*sstep*df->ySize();
		ds=(data->dataAt(0,1)-data->dataAt(0,0))*sstep*df->ySize();
		df->addRadNoise(CTFSTEPS,s0,ds,noise,0);
	}
	d=df->doIFT();
	QString s(name);
	if (forward==1) s.insert(s.length()-4,".ctf");
	else s.insert(s.length()-4,".ctfc");
	d->writeMRC(s.data());
	delete df;
	return;
}

if (forward!=-1) { printf("applyCtfTo(), forward!=-1. Unsupported.\n"); return; }
QString s(name);
s.insert(s.length()-4,".ctfc");
n=fileCount(name);
delete d;

for (i=0; i<n; i++) {
	d=new EMData;
	d->readImage(name,i);
	df=d->doFFT();
	d->gimmeFFT();
	delete d;	
	updateEqn(6,d->ySize());
	df->mult(ctfmap);
	d=df->doIFT();
	d->writeIMAGIC(s.data(),-1);
	delete df;
}

}

void CtfData::applyCtf(int noiseonly) {
float *ctf,*noise;
float s0,ds;
int i,ny,ny2;
EMData *r,*rs,*f,*r2;
char n[120];
QString nm;

if (!(Flags&CDT_HASFFT)) { warn("This image doesn't have a FFT."); return; }

ctf=data->getColumn(3)->getData();
noise=(float *)malloc(CTFSTEPS*sizeof(float));
memcpy(noise,data->getColumn(5)->getData(),CTFSTEPS*sizeof(float));
for (i=0; i<CTFSTEPS; i++) noise[i]/=10000.;

s0=data->dataAt(0,0)*sstep*fftimage->ySize();
ds=(data->dataAt(0,1)-data->dataAt(0,0))*sstep*fftimage->ySize();

r=new EMData;

if (file.findRev('.')>0) nm=file.left(file.findRev('.'));
else nm=file;
nm+=".fix";

sprintf(n,"!-%1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g %1.4g",
	defocus,exps4>.5?-B:B,amp,ampcont,noise1,noise2,noise3,noise4,V,Cs,sstep);
n[79]=0;

printf("ctf %d %f %f\n",CTFSTEPS,s0,ds);
updateEqn(2);
ny=ctfmap->ySize();

for (i=0; r->readImage(getFile(),i)==0; i++) {
	ny2=r->ySize();
	if (ny!=ny2) rs=r->clip(-(ny-ny2)/2,-(ny-ny2)/2,ny,ny);
	else rs=r->copy();
	rs->normalize();
	f=rs->doFFT();
	rs->gimmeFFT();
	delete rs;
	if (!noiseonly) f->mult(ctfmap);
	f->addRadNoise(CTFSTEPS,s0,ds,noise,0);
	r2=f->doIFT();
	f->gimmeFFT();
	r2->clearCTF();
	r2->setName(n);
	if (ny!=ny2) {
		rs=r2->clip((ny-ny2)/2,(ny-ny2)/2,ny2,ny2);
		rs->writeIMAGIC(nm,i);
		delete rs;
	}
	else r2->writeIMAGIC(nm,i);
	delete f;
	delete r2;
	printf("%d done\n",i);
}
free(noise);
}

// This is for real-time update of one image
void CtfData::applyCtf(EMData *im) {
float *ctf,*noise;
float s0,ds;
int i,ny,ny2;
EMData *r,*rs,*f,*r2;
char n[120];
QString nm;

if (!(Flags&CDT_HASFFT)) { warn("This image doesn't have a FFT."); return; }

ctf=data->getColumn(3)->getData();
noise=(float *)malloc(CTFSTEPS*sizeof(float));
memcpy(noise,data->getColumn(5)->getData(),CTFSTEPS*sizeof(float));
for (i=0; i<CTFSTEPS; i++) noise[i]/=10000.;

s0=data->dataAt(0,0)*sstep*fftimage->ySize();
ds=(data->dataAt(0,1)-data->dataAt(0,0))*sstep*fftimage->ySize();

//recopy parent
if (im->Parent()) {
	im->setRAlign(0.);
	im->setTAlign(0.,0.);
	im->rotateAndTranslateFast();
}
else {
	printf("No parent (%p)\n",im->Parent());
	im->setParent(im->copy(0,0));
}

updateEqn(2);
ny=ctfmap->ySize();

	ny2=im->ySize();
	if (ny!=ny2) rs=im->clip(-(ny-ny2)/2,-(ny-ny2)/2,ny,ny);
	else rs=im->copy();
	rs->normalize();
	f=rs->doFFT();
	rs->gimmeFFT();
	delete rs;
	f->mult(ctfmap);
	f->addRadNoise(CTFSTEPS,s0,ds,noise,0);
	r2=f->doIFT();
	f->gimmeFFT();

	float *dd=im->getData();
	float *dst=r2->getDataRO();
	memcpy(dd,dst,im->xSize()*im->ySize()*sizeof(float));
	im->doneData();
	r2->doneData();

	delete f;
	delete r2;

free(noise);
}


void CtfData::updateImage() {
float *dat;
int i;
float s0,ds;
if (!data) return;

data->getColumn(0)->setRange(smax/(float)CTFSTEPS,smax,CTFSTEPS);
if (!fftimage) return;

s0=data->dataAt(0,0)*sstep*fftimage->ySize();
ds=(data->dataAt(0,1)-data->dataAt(0,0))*sstep*fftimage->ySize();
//printf("updateimage %1.2f %1.2f\n",s0,ds);
//printf("%f %f\n",s0,ds);
dat=data->getColumn(10)->getData();
if (stigdf>-.01) fftimage->calcRadDist(CTFSTEPS,s0,ds,dat);
else fftimage->calcRadDist(CTFSTEPS,s0,ds,dat,stigang*PI/180.0,10.0*PI/180.0);
//for (i=0; i<CTFSTEPS; i++) dat[i]*=dat[i];
//dat[0]=0;
data->getColumn(10)->setData();

dat=data->getColumn(12)->getData();
if (stigdf==0) fftimage->calcRadDist(CTFSTEPS,s0,ds,dat);
else fftimage->calcRadDist(CTFSTEPS,s0,ds,dat,(stigang-90.0)*PI/180.0,10.0*PI/180.0);
//for (i=0; i<CTFSTEPS; i++) dat[i]*=dat[i];
//dat[0]=0;
data->getColumn(12)->setData();

}

void CtfData::warn(const char *msg) {
	QMessageBox::warning(NULL,"CTFIT Warning",msg,"Ok");
	return;
}

void CtfData::copy(CtfData *src) {
if (!src) return;
setStrucParm(src->cfilmsf);
setStrucMode(src->structype,src->strucfile);
setStigParm(src->stigang,src->stigdf);
setNoiseParm(src->noise1,src->noise2,src->noise3,src->noise4);
setSigParm(src->defocus,src->B,src->amp,src->ampcont,src->exps4);
setImParm(src->V,src->Cs,src->sstep,src->smax);

updateEqn();
}

// makemap  1->phase flip map, 2->2d Fourier apply CTF, 6-> 2d Fourier fix CTF
// 3->3d Fourier apply CTP, 4->3d Fourier fix CTF, 5->set x-ray SF to 3d
void CtfData::updateEqn(int makemap,int size) {

if (!data) return;
if (ampcont<0||ampcont>1.0) { ampcont=0.0; printf("Warning, ampcont out of range\n"); }
if (V<0) V=100.0;

lambda=12.2639/sqrt(V*1000.0+.97845*V*V);
zero=sqrt(-(defocus*10000.0+sqrt(defocus*defocus*1.0e8-2.0*lambda*Cs*1.0e7+
	2.0*lambda*Cs*1.0e7/PI*asin(ampcont)))/(lambda*lambda*Cs*1.0e7));

zerostig=sqrt(-((defocus+stigdf)*10000.0+sqrt(SQR(defocus+stigdf)*1.0e8-2.0*lambda*Cs*1.0e7+
	2.0*lambda*Cs*1.0e7/PI*asin(ampcont)))/(lambda*lambda*Cs*1.0e7));

data->setVar("V",V,TRUE);
data->setVar("Cs",Cs,TRUE);
data->setVar("lambda",lambda,TRUE);
data->setVar("stigdf",stigdf,TRUE);
data->setVar("defocus",defocus,TRUE);
data->setVar("exps4",exps4,TRUE);
data->setVar("amp",amp,TRUE);
data->setVar("ampcont",ampcont,TRUE);
data->setVar("B",B,TRUE);
data->setVar("noise4",noise4,TRUE);
data->setVar("noise1",noise1,TRUE);
data->setVar("noise2",noise2,TRUE);
data->setVar("noise3",noise3,TRUE);
data->setVar("cfilmsf",cfilmsf,TRUE);
data->setVar("Q1",Q[0],TRUE);
data->setVar("Q2",Q[1],TRUE);
data->setVar("Q3",Q[2],TRUE);
data->setVar("Q4",Q[3],TRUE);
data->setVar("Q5",Q[4],TRUE);
data->setVar("Q6",Q[5]);

// The new background mode requires some nonfunctional calculation
if (bgmode) {
	// First we find the x,y positions of the zeros (to the nearest pixel)
	int n=1;
	float x[50],y[50];
	int i;
	float lv=1.0;
	for (i=0; i<CTFSTEPS; i++) {
		float s=(i+1)*smax/(float)CTFSTEPS;
		float gamma=-2.*PI*(Cs*1.0e7/4.*pow(lambda,3.0F)*pow(s,4.0F)+defocus*5000.*lambda*s*s);
		float v=(sqrt(1.-ampcont*ampcont)*sin(gamma)+ampcont*cos(gamma));
		if ((v*lv<0 || v==0) && n<49) {
			x[n]=i;
			y[n]=data->dataAt(10,i)-(data->dataAt(13,i)+cfilmsf*data->dataAt(15,i))*pow(data->dataAt(3,i),2.0F);
			n++;
		}
		lv=v;
	}
	x[0]=0;
	y[0]=noise1;
	x[n]=CTFSTEPS-1;
	y[n]=data->dataAt(10,x[n]);
	n++;

	// linear interpolation between zeros for the moment
	float *d=data->getColumn(5)->getData();
	for (i=0; i<CTFSTEPS; i++) {
		d[i]=interpolate(i,n,x,y);
	}

	// See if there are nearby lower points
	for (i=1; i<n-1; i++) {
		if (x[i-1]!=x[i]-1 && d[(int)x[i]-1]>data->dataAt(10,x[i]-1)) { x[i]--; y[i]=data->dataAt(10,x[i]); }
		if (x[i+1]!=x[i]-1 && d[(int)x[i]+1]>data->dataAt(10,x[i]+1)) { x[i]++; y[i]=data->dataAt(10,x[i]); }
	}

	// sqrt interpolation
	for (i=0; i<n; i++) y[i]=(y[i]<=0)?0.:sqrt(y[i]);
	for (i=0; i<CTFSTEPS; i++) {
		d[i]=pow(interpolate(i,n,x,y),2.0F);
	}
	
	data->getColumn(5)->setData(d);
}


if (!fftimage && makemap<3) return;	// no point in going on if there isn't any data

// make a sign map which can be multiplied by a fourier image to correct phase flips
if (makemap==1) {
	int i,j,ny;
	float s,ds,*d;
	
	if (ctfmap) delete ctfmap;
	ctfmap=fftimage->copyHead();
	ctfmap->setRI(1);

	cmgamma->setVar("Q1",Q[0],TRUE);
	cmgamma->setVar("Q2",Q[1],TRUE);
	cmgamma->setVar("Q3",Q[2],TRUE);
	cmgamma->setVar("Q4",Q[3],TRUE);
	cmgamma->setVar("Q5",Q[4],TRUE);
	cmgamma->setVar("Q6",Q[5],TRUE);
	cmgamma->setVar("Cs",Cs,TRUE);
	cmgamma->setVar("lambda",lambda,TRUE);
	cmgamma->setVar("defocus",defocus);	

	cmctf->setVar("Q1",Q[0],TRUE);
	cmctf->setVar("Q2",Q[1],TRUE);
	cmctf->setVar("Q3",Q[2],TRUE);
	cmctf->setVar("Q4",Q[3],TRUE);
	cmctf->setVar("Q5",Q[4],TRUE);
	cmctf->setVar("Q6",Q[5],TRUE);
	cmctf->setVar("B",B,TRUE);
	cmctf->setVar("exps4",exps4,TRUE);
	cmctf->setVar("ampcont",ampcont,TRUE);
	cmctf->setVar("amp",amp);

	ds=1.0/(sstep*ctfmap->ySize());
	
	ny=ctfmap->ySize();
	d=ctfmap->getData();

	for (j=0; j<ny; j++) {
		for (i=0; i<ny+2; i+=2) {
			s=hypot((float)i/2.0,(float)j-ny/2.0)*ds;
			cmgamma->setVar("S",s);
			cmctf->setVar("gamma",cmgamma->eval(),TRUE);
			cmctf->setVar("S",s);
			if (cmctf->eval()>0) { *d=1.0; d++; *d=1.0; d++; } 
			else { *d=-1.0; d++; *d=-1.0; d++; }
		}
	}
	
	ctfmap->doneData();
	
	ctfmap->writeMRC("signmap.mrc");
}


// make a Wiener map which can be multiplied by a fourier image to correct phase flips
if (makemap==11) {
	int i,j,ny;
	float s,ds,*d;

	int is;
	float SNR,noiseInt,lessthan1Factor;
	float deltaf = defocus*(1.e-6);
	float epsilon = asin((ampcont));
	float lam = lambda*1.e-10;
	float C_s = Cs*1.e-3;

	float FirstZero =  sqrt(lam)* sqrt(       \
		(sqrt( deltaf*deltaf - 2*lam *C_s* ( 1 - epsilon) )  \
		- deltaf )/(2 *( 1 - epsilon) )); 
	
	float spacing = (data->dataAt(0,2)) - (data->dataAt(0,1));
	int IFirstZero = 1./(FirstZero*1.e10*spacing);
	float FirstZeroAmp = data->dataAt(13,IFirstZero);
	float fixtmp, CTFStuff, IntensityStuff, SFStuff;
	
	if (ctfmap) delete ctfmap;
	ctfmap=fftimage->copyHead();
	ctfmap->setRI(1);

	cmgamma->setVar("Q1",Q[0],TRUE);
	cmgamma->setVar("Q2",Q[1],TRUE);
	cmgamma->setVar("Q3",Q[2],TRUE);
	cmgamma->setVar("Q4",Q[3],TRUE);
	cmgamma->setVar("Q5",Q[4],TRUE);
	cmgamma->setVar("Q6",Q[5],TRUE);
	cmgamma->setVar("Cs",Cs,TRUE);
	cmgamma->setVar("lambda",lambda,TRUE);
	cmgamma->setVar("defocus",defocus);	

	cmctf->setVar("Q1",Q[0],TRUE);
	cmctf->setVar("Q2",Q[1],TRUE);
	cmctf->setVar("Q3",Q[2],TRUE);
	cmctf->setVar("Q4",Q[3],TRUE);
	cmctf->setVar("Q5",Q[4],TRUE);
	cmctf->setVar("Q6",Q[5],TRUE);
	cmctf->setVar("B",B,TRUE);
	cmctf->setVar("exps4",exps4,TRUE);
	cmctf->setVar("ampcont",ampcont,TRUE);
	cmctf->setVar("amp",amp);

	ds=1.0/(sstep*ctfmap->ySize());
	
	ny=ctfmap->ySize();
	d=ctfmap->getData();

	int iquest=4;
	float	fudgeconstant = .3;
	for (j=0; j<ny; j++) {
		for (i=0; i<ny+2; i+=2) {
//			s=hypot((float)i/2.0,(float)j-ny/2.0)*ds;
			s=hypot((float)i/2.0,(float)j-ny/2.0)*2.;
			cmgamma->setVar("S",s);
			cmctf->setVar("gamma",cmgamma->eval(),TRUE);
			cmctf->setVar("S",s);
			is = s ;
//			if (cmctf->eval()>0) { *d=1.0; d++; *d=1.0; d++; }  
//			else { *d=-1.0; d++; *d=-1.0; d++; } USED to be for  makemap=1
			fixtmp =0;
			if (is<ny-1) {
//	   			if (data->dataAt(10,is)>0) 


					CTFStuff = (s-is)* (data->dataAt(3,is))+ (is+1-s)*(data->dataAt(3,is+1));					
					noiseInt =  (s-is)* (data->dataAt(5,is))+ (is+1-s)*(data->dataAt(5,is+1));

					if (1) {        // From Xray
						SFStuff = (s-is)*(data ->dataAt(13,is)) +  
                                    (is+1-s)*(data->dataAt(13,is+1));   
						SNR =  (CTFStuff* SFStuff) * (CTFStuff* SFStuff) / noiseInt;					
						lessthan1Factor = 2.* sqrt(SNR)/(fudgeconstant + SNR);
 					// lessthan1Factor = 2.* 1/(sqrt(1.+SNR));	for SF data
					} else{         // From raw data
						IntensityStuff	= (s-is)*(data ->dataAt(10,is)) + 
										(is+1-s)*(data ->dataAt(10,is+1));	
						if (IntensityStuff<noiseInt)  IntensityStuff = 1.001*noiseInt;			
						SFStuff =  (s-is)* (data->dataAt(11,is))+ 
									(is+1-s)*(data->dataAt(11,is+1));						
						SFStuff =   sqrt(IntensityStuff-noiseInt) ;
						lessthan1Factor = 2.* sqrt(noiseInt)*CTFStuff*SFStuff/
						  (sqrt(CTFStuff*CTFStuff)*IntensityStuff); 
					}

					fixtmp =  SFStuff * lessthan1Factor / (2.*sqrt(noiseInt)) ;
//					fixtmp = 4.;
					if ( CTFStuff <0) fixtmp *= -1;
				if (j==ny/2) { printf("%d,%d,%d, %f %f	%f	%f\n", i, j, 
				 						is, fixtmp, SFStuff, noiseInt, lessthan1Factor);}
			}
//			if (is < IFirstZero) fixtmp *= (sqrt(FirstZeroAmp /(data->dataAt(13,is)
//			)));
			*d = fixtmp; d++; *d = fixtmp; d++;
		}
	}
	ctfmap ->  filter(0,2,1);
	ctfmap ->  radialAverage();
	ctfmap->doneData();
	
	ctfmap->writeMRC("signmap.mrc");
}


// make a map which can be multiplied by a fourier image to apply a CTF
if (makemap==2 || makemap==6) {
	int i,j,ny;
	float s,ds,*d,e;
	
	if (ctfmap) delete ctfmap;
	ctfmap=fftimage->copyHead();
	ctfmap->setRI(1);

	cmgamma->setVar("Q1",Q[0],TRUE);
	cmgamma->setVar("Q2",Q[1],TRUE);
	cmgamma->setVar("Q3",Q[2],TRUE);
	cmgamma->setVar("Q4",Q[3],TRUE);
	cmgamma->setVar("Q5",Q[4],TRUE);
	cmgamma->setVar("Q6",Q[5],TRUE);
	cmgamma->setVar("Cs",Cs,TRUE);
	cmgamma->setVar("lambda",lambda,TRUE);
	cmgamma->setVar("defocus",defocus);	

	cmctf->setVar("Q1",Q[0],TRUE);
	cmctf->setVar("Q2",Q[1],TRUE);
	cmctf->setVar("Q3",Q[2],TRUE);
	cmctf->setVar("Q4",Q[3],TRUE);
	cmctf->setVar("Q5",Q[4],TRUE);
	cmctf->setVar("Q6",Q[5],TRUE);
	cmctf->setVar("B",B,TRUE);
	cmctf->setVar("exps4",exps4,TRUE);
	cmctf->setVar("ampcont",ampcont,TRUE);
	cmctf->setVar("amp",amp);

	ds=1.0/(sstep*ctfmap->ySize());
	
	ny=ctfmap->ySize();
	d=ctfmap->getData();

	for (j=0; j<ny; j++) {
		for (i=0; i<ny+2; i+=2) {
			s=hypot((float)i/2.0,(float)j-ny/2.0)*ds;
			cmgamma->setVar("S",s);
			cmctf->setVar("gamma",cmgamma->eval(),TRUE);
			cmctf->setVar("S",s);
			e=cmctf->eval();
			if (makemap==6) {
				if (fabs(e)<.04) e=25; else e=1.0/e;
			}
			*d=e;
			d++;
			*d=e;
			d++;
		}
	}
	
	ctfmap->doneData();
	
	ctfmap->writeMRC("ctfmap.mrc");
}

// this makes a 3d amp mult map for modifying 3d volumes
// 3 adds a CTF, 4 does a simple CTF correction
if (makemap==3 || makemap==4) {
	int i,j,k,f=0,ny;
	float s,ds,*d,e;
	
	if (ctfmap) delete ctfmap;
	ctfmap=new EMData;
	ctfmap->setSize(size+2,size,size);
	ctfmap->setRI(1);

	cmgamma->setVar("Q1",Q[0],TRUE);
	cmgamma->setVar("Q2",Q[1],TRUE);
	cmgamma->setVar("Q3",Q[2],TRUE);
	cmgamma->setVar("Q4",Q[3],TRUE);
	cmgamma->setVar("Q5",Q[4],TRUE);
	cmgamma->setVar("Q6",Q[5],TRUE);
	cmgamma->setVar("Cs",Cs,TRUE);
	cmgamma->setVar("lambda",lambda,TRUE);
	cmgamma->setVar("defocus",defocus);	

	cmctf->setVar("Q1",Q[0],TRUE);
	cmctf->setVar("Q2",Q[1],TRUE);
	cmctf->setVar("Q3",Q[2],TRUE);
	cmctf->setVar("Q4",Q[3],TRUE);
	cmctf->setVar("Q5",Q[4],TRUE);
	cmctf->setVar("Q6",Q[5],TRUE);
	cmctf->setVar("B",B,TRUE);
	cmctf->setVar("exps4",exps4,TRUE);
	cmctf->setVar("ampcont",ampcont,TRUE);
	cmctf->setVar("amp",amp);

	ds=1.0/(sstep*ctfmap->ySize());
	
	ny=ctfmap->ySize();
	d=ctfmap->getData();

	for (k=0; k<ny; k++) {
		for (j=0; j<ny; j++) {
			for (i=0; i<ny+2; i+=2) {
				s=sqrt(i*i/4.0+(j-ny/2)*(j-ny/2)+(k-ny/2)*(k-ny/2))*ds;
				cmgamma->setVar("S",s);
				cmctf->setVar("gamma",cmgamma->eval(),TRUE);
				cmctf->setVar("S",s);
				e=cmctf->eval();
				// 1/ctf, but stop at 1st zero
				if (makemap==4) {
/*					if (fabs(e)>.05 && !f) f=1;
					if (fabs(e)<.05 && f) f=2; 
					if (f==2) e=20.0; else e=1.0/e; */
					if (fabs(e)<.04) e=25; else e=1.0/e;
				}
				*d=e;
				d++;
				*d=e;
				d++;
			}
		}
	}
	
	ctfmap->doneData();
	
//	ctfmap->writeMRC("ctfmap.mrc");

}

// This will make a 3d amp mult map which allows an x-ray SF to be
// applied to a 3d data set
if (makemap==5) {
	int i,j,k,r,ny;
	float s,ds,*d,e,l=1.0;
	
	if (ctfmap) delete ctfmap;
	ctfmap=new EMData;
	ctfmap->setSize(size+2,size,size);
	ctfmap->setRI(1);

	cmgamma->setVar("Q1",Q[0],TRUE);
	cmgamma->setVar("Q2",Q[1],TRUE);
	cmgamma->setVar("Q3",Q[2],TRUE);
	cmgamma->setVar("Q4",Q[3],TRUE);
	cmgamma->setVar("Q5",Q[4],TRUE);
	cmgamma->setVar("Q6",Q[5],TRUE);
	cmgamma->setVar("Cs",Cs,TRUE);
	cmgamma->setVar("lambda",lambda,TRUE);
	cmgamma->setVar("defocus",defocus);	

	cmctf->setVar("Q1",Q[0],TRUE);
	cmctf->setVar("Q2",Q[1],TRUE);
	cmctf->setVar("Q3",Q[2],TRUE);
	cmctf->setVar("Q4",Q[3],TRUE);
	cmctf->setVar("Q5",Q[4],TRUE);
	cmctf->setVar("Q6",Q[5],TRUE);
	cmctf->setVar("B",B,TRUE);
	cmctf->setVar("exps4",exps4,TRUE);
	cmctf->setVar("ampcont",ampcont,TRUE);
	cmctf->setVar("amp",amp);

	ds=1.0/(sstep*ctfmap->ySize());
	
	ny=ctfmap->ySize();
	d=ctfmap->getData();

	for (k=0; k<ny; k++) {
		for (j=0; j<ny; j++) {
			for (i=0; i<ny+2; i+=2) {
				s=sqrt(i*i/4.0+(j-ny/2)*(j-ny/2)+(k-ny/2)*(k-ny/2))*ds/(smax/CTFSTEPS)-1.0;
				r=fabs(s);	// r is the element number in the data array
				s-=r;		// s is now remainder for interpolation
				
				if (data->dataAt(10,r+1)<=0||data->dataAt(10,r+1)<=0||data->dataAt(13,r+1)==1.0) e=l;
				else {
					e=s*data->dataAt(13,r+1)/data->dataAt(10,r+1)+
						(1.0-s)*data->dataAt(13,r)/data->dataAt(10,r);
					if (e<0) e=0; else e=sqrt(e);
					l=e;
				}
				
				*d=e;
				d++;
				*d=e;
				d++;
			}
		}
	}
	
	ctfmap->doneData();
	
//	ctfmap->writeMRC("ctfmap.mrc");

}

}
