#include <stdio.h>
#include <math.h>
#include "util.h"
#include "Vect.h"
#include "Euler.h"
#include <string.h>

//#define DEBUG0
//
//#define PI 3.1415926
//       %% Euler.C         %%                    Philip R. Baldwin 11/5/98
//            This is a C++ class,
//               for performing operations on Euler angles and rotations.
//             The angles are kept internally in EMAN format (altitude,
//              azimuth, phi) or (theta, psi, phi) following Goldstein's
//                 mechanics book p.147.
//             The rotation angles may be accessed in various electrom
//               microscopy formats  and intraconverted.
//             In addition symmetry operations are created for various
//                  common crystal types: thus from an initial angle
//                  assignment, symmetry equivalent angles may be created.
//             The private variables may be viewed from Euler.h
//                   they are
//						char *symmetry_;  int symcounter_;
//                      int typ_; float alt_ , az_ , phi_ ;
//                      Euler * symoperation_; int MaxSymEl_;
//
//               Many of the functions are overloaded,
//                 since rotations may be defined using quaternions, points
//                  on S^3,  three by three matrices, spin- and spin-axis
//                  and IMAGIC format.


//       Arguments in Radians!!!!!!!!!!!!!!!!!!!!!!

//      THE CONSTRUCTORS
Euler::Euler() { 
alt_=0.; az_=0.; phi_=0.;
init();
} // already rectified

	
Euler::Euler(float alt, float az, float phi){// EMAN by default
	setAngle(alt,az,phi);
	init();
}

Euler::Euler(float a1, float a2, float a3, int typ) { 
	setAngle(a1,a2,a3,typ);
	init();
} // end EMAN, IMAGIC, SPIDER constructor

Euler::Euler(float e1, float e2, float e3, float e0, int typ) {
	setAngle(e1,e2,e3,e0,typ); 
	init();
}           // end quaternion, and spin constructor	

Euler::Euler(float *m){
	setAngle(m);
	init();
}

void Euler::init() {
strcpy(sym,"i");
nsym=1;
csym=0;
}

Euler::~Euler() {
}

void Euler::rectify(){
#ifdef DEBUG0
	printf("rectify %f,%f,%f ->",alt_,az_,phi_);
#endif
	if (!goodf(&phi_)) phi_=0;
	if (!goodf(&alt_)) alt_=0;
	if (!goodf(&az_)) az_=0;
	if (sqrt(alt_*alt_) <= .0001) { az_ += phi_; phi_=0;}
	
	phi_=fmod(phi_,(float)(2.0*PI));
	if (phi_<0) phi_+=2.0*PI;
	
	az_=fmod(az_,(float)(2.0*PI));
	if (az_<0) az_+=2.0*PI;
#ifdef DEBUG0
	printf(" %f,%f,%f\n",alt_,az_,phi_);
#endif
}

/* *****************************************************************************  */
//		               FUNCTIONS
//   ********   1. Set Type  EMAN, IMAGIC, QUATernion, SPIN
//     

void Euler::setTyp(int typ) { typ_=typ ;}


//          2.    setAngle  overloaded( 
//
	
void Euler::setAngle(float a1, float a2, float a3, int typ){ 
#ifdef DEBUG0
	printf("setAngle %f %f %f %d\n",a1,a2,a3,typ);
#endif
	alt_ =a1 ; az_ = a2 ; phi_ = a3;
	if (typ == IMAGIC) { /*az_ +=PI; phi_ +=PI;*/} // then a1= beta, a2=gamma, a3=alpha
	if (typ == SPIDER) { az_ = a2-PI/2.-PI; phi_ = a3+PI/2.-PI;}    // then a1=theta, a2=phi, a3=gamma
    	if (typ !=EMAN && typ !=IMAGIC && typ != SPIDER) {	 
			printf("unknown declaration of type for Euler constructor /n");} 
			rectify();
}           // end EMAN and IMAGIC setAngle
			
void  Euler::setAngle(float e1, float e2, float e3, float e0, int typ) {
	typ_ = typ ;
	if (typ !=SPIN && typ!=QUAT && typ != SGIROT) {	 
		printf("unknown declaration of type for Euler constructor /n");}
//   if not already, put in S4 language
	if (typ == SPIN || typ==SGIROT) { 
		float normal = sqrt(e1*e1+e2*e2+e3*e3);
		float sinQo2 = sin(e0/2.);		e0 = cos(e0/2.);
		e1 *= sinQo2/normal; e2 *= sinQo2/normal; e3 *= sinQo2/normal; }
	float w = e0, x = e1, y = e2, z = e3;
	m_[0] = 1-2*y*y-2*z*z;;
	m_[1] = 2*x*y + 2*w*z;
	m_[2] = 2*x*z - 2*w*y;
	m_[3] = 2*x*y - 2*w*z;
	m_[4] = 1-2*x*x-2*z*z;
	m_[5] = 2*y*z + 2*w*x;
	m_[6] = 2*x*z + 2*w*y;
	m_[7] = 2*y*z - 2*w*x;
	m_[8] = 1 - 2*x*x - 2*y*y;
	
	setAngle(m_);
	
	//alt_ = acos((e0*e0 - e1*e1 - e2*e2 + e3*e3)/(e0*e0+e1*e1+e2*e2+e3*e3)); 
	//az_  = atan2( e2*e0 - e1*e3 , e1*e0 + e2*e3); 
	//phi_ = atan2(-e2*e0 - e1*e3 , e1*e0 - e2*e3); 
	rectify();
	if (typ==SGIROT) inverse();
}           // end quaternion, and spin setAngle

void Euler::setAngle(float *m){
	typ_ = MATRIX;
	alt_ = acos( m[8]    );
	az_  = atan2(m[6],-m[7]);
	phi_ = atan2(m[2], m[5]);
	if (m[8] > 0.999999) {
		alt_=0;
		az_ = atan2(m[1],m[4]);
		phi_ =0;}
	if (m[8] < -0.999999) {
		alt_ = PI;
		az_ = atan2(m[1],-m[4]);
		phi_ =0;
	}		 		 
	rectify();
}
// ********** %% 2.3 Increment an Euler %%
void Euler::increment(float delAlt, float delAz, float delPhi){
	alt_ += delAlt; az_ += delAz; phi_ += delPhi; rectify();}

float Euler::diff(Euler* ang, bool nophi){
float a[3], b[3],p,ret;
if(nophi){
	a[0]=sin(alt())*cos(az());	a[1]=sin(alt())*sin(az());	a[2]=cos(alt());
	b[0]=sin(ang->alt())*cos(ang->az());	b[1]=sin(ang->alt())*sin(ang->az());  b[2]=cos(ang->alt());
	//p=fabs(a[0]*b[0]+a[1]*b[1]+a[2]*b[2]);
	p=(a[0]*b[0]+a[1]*b[1]+a[2]*b[2]);
} else {
	float *m=mat(), *m2=ang->mat();
	float denom=sqrt(3.0);	
	a[0]=(m[0]+m[1]+m[2])/denom;	a[1]=(m[3]+m[4]+m[5])/denom;	a[2]=(m[6]+m[7]+m[8])/denom;	
	b[0]=(m2[0]+m2[1]+m2[2])/denom;	b[1]=(m2[3]+m2[4]+m2[5])/denom;	b[2]=(m2[6]+m2[7]+m2[8])/denom;	
	//p=fabs((a[0]*b[0]+a[1]*b[1]+a[2]*b[2]));
	p=((a[0]*b[0]+a[1]*b[1]+a[2]*b[2]));
}
if(p>1)	p=1; else if (p<-1)p=-1;
ret=acos(p);
//printf("a=(%f,%f,%f)\tb=(%f,%f,%f)\ta.b=%f\tang=%f\n",a[0],a[1],a[2],b[0],b[1],b[2],p,ret*180/PI);
return ret;
}		

//    ******** %% 2.5  Take an inverse of an Euler %%
Euler Euler::inverse() {
        Euler ereturn(alt_,3.*PI - phi_,3.*PI- az_ );
       ereturn.rectify();  
        return ereturn;
}

//    ******** %% 3.  Multiplying Eulers (from Left) %%

void Euler::fromLeft(Euler leftEuler) {
//	find ac = bc* ab   = leftEuler * currentEuler
	float * mreturn =matmulmat(leftEuler.mat(), (*this).mat());
	Euler ereturn( mreturn );
	alt_ = ereturn.alt(); az_= ereturn.az(); phi_ = ereturn.phi();
	free(mreturn);
}  // end applying Euler fromLeft




void Euler::fromLeft(float alt, float az, float phi) {
//	find ac = bc* ab   = leftEuler * currentEuler
	Euler ereturn(alt,az,phi);
	float * mreturn =matmulmat(ereturn.mat(), (*this).mat());
	ereturn.setAngle( mreturn );
	alt_ = ereturn.alt(); az_= ereturn.az(); phi_ = ereturn.phi();
	free(mreturn);
}  // end applying Euler fromLeft



Euler Euler::EulerMult(Euler bc) {
	Euler ac(alt_, az_ , phi_);  // actually this starts ac at ab
	 ac.fromLeft(bc);		   // but now it rotates ac to
	 return ac;         // the value of interest
} // end Euler multiply

Euler Euler::operator *(Euler ab) {
	 ab.fromLeft(*this); 
	 return ab;                           
} // end Euler multiply

Euler Euler::operator /(Euler ab) {
	return ((*this) * (ab.inverse()));
} // end Euler divide

int Euler::operator !=(Euler ab) {
	int iquestalt = (ab.alt() != (*this).alt());
	int iquestaz = (ab.az() !=   (*this).az());
	int iquestphi = (ab.phi() != (*this).phi());
	return iquestalt+iquestaz+iquestphi;
}

//    ***     %% 4.  SYMMETRY   SYMMETRY SYMMERY  %% !!!!!!!!!!!!!!!!!!!!!!!!!
// **********     ***********

int Euler::getMaxSymEl() {
return nsym;
}

char* Euler::getSym() { return sym; }

void Euler::setSym(char *symmetry) {
if (!symmetry) { strcpy(sym,"i"); return; }
strncpy(sym,symmetry,7);
for (int i=0; i<7; i++) if (isalpha(sym[i])) sym[i]=tolower(sym[i]);
sym[7]=0;
csym=0;
if (tolower(sym[0])=='c') nsym=atoi(sym+1);
else if (sym[0]=='d') nsym=atoi(sym+1)*2;
else if (strcmp(sym,"icos")==0) nsym=60;
else if (strcmp(sym,"oct")==0) nsym=24;
else if (strcmp(sym,"tet")==0) nsym=12;
else if (strcmp(sym,"i")==0) nsym=1;
else nsym=0;
}

// returns the n'th symmetric element for the current symmetry
Euler Euler::SymN(int n) {
static Euler invalid(-.1,-.1,-.1);
Euler ret;

static float ICOS[180]={
	0.,0.  ,   0., 0.,0. ,  288., 0.,0. ,  216., 0.,0. ,  144., 0.,0. ,   72., \
	0.,63.4349 , 36., 0.,63.4349 , 324., 0.,63.4349 , 252., 0.,63.4349 , 180., 0.,63.4349 , 108., \
	72.,63.4349 , 36., 72.,63.4349 , 324., 72.,63.4349 , 252., 72.,63.4349 , 180., 72.,63.4349 , 108.,\
	144.,63.4349,36.,144.,63.4349 ,324.,144.,63.4349 , 252.,144.,63.4349 , 180., 144.,63.4349, 108., \
	216.,63.4349 , 36., 216.,63.4349,324., 216.,63.4349 ,252.,216.,63.4349 ,180., 216.,63.4349 ,108., \
	288.,63.4349 , 36., 288.,63.4349,324., 288.,63.4349,252., 288.,63.4349,180.,288.,63.4349 ,108.,\
	36.,116.5651 ,0., 36.,116.5651,288., 36.,116.5651 ,216., 36.,116.5651 ,144., 36.,116.5651 ,72.,\
	108.,116.5651 ,0.,108.,116.5651 ,288., 108.,116.5651 ,216.,108.,116.5651 ,144., 108.,116.5651 ,72.,\
	180.,116.5651 ,0.,180.,116.5651 ,288., 180.,116.5651 ,216., 180.,116.5651 ,144., 180.,116.5651 ,72.,\
	252.,116.5651 ,0., 252.,116.5651 ,288., 252.,116.5651 ,216., 252.,116.5651 ,144., 252.,116.5651 ,72.,\
	324.,116.5651 ,0.,324.,116.5651 ,288., 324.,116.5651 ,216., 324.,116.5651 ,144., 324.,116.5651 ,72.,\
	0.,180.,0., 0.,180.,288., 0.,180.,216., 0.,180.,144., 0.,180.,72.  
};
float OCT[24*3]={
	 0, 0, 0,   90, 0, 0,   180, 0, 0,   270,0, 0,   90,90, 0,   90,270, 0,
	 0, 0,90,   90, 0,90,   180, 0,90,   270,0,90,   90,90,90,   90,270,90,
	 0, 0,180,   90, 0,180,   180, 0,180,   270,0,180,   90,90,180,   90,270,180,
	 0, 0,270,   90, 0,270,   180, 0,270,   270,0,270,   90,90,270,   90,270,270
};

// eman2 convention with the face along z-axis
// the implementation is really dirty at the moment as I didn't
// take symmetry course during my Ph.D. Hmm. Time to read.
// Teepanis.
#define lvl1 109.4712f
float TET[12*3]={
          0,0,0,   0,0,120,    0,0,240,
          0,lvl1,60,   0,lvl1,180,    0,lvl1,300,
        120,lvl1,60, 120,lvl1,180,  120,lvl1,300,
        240,lvl1,60, 240,lvl1,180,  240,lvl1,300
};
#undef lvl1
if (n>=nsym) return invalid;

if (tolower(sym[0])=='c') ret.setAngle(0,n*2.0*PI/nsym,0);
else if (sym[0]=='d') {
	if (n>=nsym/2) ret.setAngle(PI,(n-nsym/2)*2.0*PI/(nsym/2),0);
	else ret.setAngle(0,n*2.0*PI/(nsym/2),0);
}
else if (strcmp(sym,"icos")==0) ret.setAngle(ICOS[n*3+1]*PI/180.0,ICOS[n*3+2]*PI/180.0-PI/2.,ICOS[n*3]*PI/180.0+PI/2.);
else if (strcmp(sym,"oct")==0) ret.setAngle(OCT[n*3]*PI/180.0,OCT[n*3+1]*PI/180.0,OCT[n*3+2]*PI/180.0);
else if (strcmp(sym,"tet")==0) ret.setAngle(TET[n*3+1]*PI/180.0,TET[n*3+2]*PI/180.0-PI/2.,TET[n*3]*PI/180.0+PI/2.);
else if (strcmp(sym,"i")==0) ret.setAngle(0,0,0);

return (*this*ret);
}

void Euler::resetSym(){
csym=0;
return;	
}

Euler Euler::FirstSym(){
resetSym();
return NextSym();
}

Euler Euler::NextSym() {
	csym++;
	return SymN(csym-1);
}  // end NextSym

int Euler::valid() {
	if (alt() < 0) return 0;
	return 1;
}

vector< Euler*> Euler::asymmetricUnit(float angle_step, int mode, int with_mirror) // return the complete list of orientations in an asymmetric unit
{
	if (angle_step<=0) {
		fprintf(stderr, "ERROR: Euler::asymmetricUnit(float angle_step, int mode) angle_step=%f\n",angle_step);
		exit(1);
	}
	vector< Euler* > ret;
	float az, az0=0, az1=2.0*PI;
	float alt, alt0=0, alt1=PI/2.0;
	float mirror_factor=2.0;
	int j, csym;
	
	if (with_mirror) mirror_factor = 1.0;
	
	if (tolower(sym[0])=='c' || tolower(sym[0])=='d' ) {
		csym=atoi(&sym[1]);
		az1/=(float)csym;
		if (tolower(sym[0])=='d') az1/=2.0;
		alt1=PI/mirror_factor;
	}
	else if (strncmp(sym,"icos",4)==0) {	// seems too large region
		csym=60;
		az1/=5.0*mirror_factor;
		alt1=37.3773681406497*PI/180.0; // 5fold axis to the closest 3fold axis
	}
	else if (strncmp(sym,"oct",3)==0) {		// is this right ?
		csym=24;
		az1/=4.0;
		alt1=54.736*PI/180.0;
	}
	
	if (csym==60) csym=-5;
	if (csym==24) csym=-4;
	
	if(mode == 0) { // "prop" mode as in project3d
		int i=0;
		for (alt=alt0, j=0; alt<=alt1+angle_step*PI/180.0/2; alt+=angle_step*PI/180.0, j++) {
			float h=floor(360.0/(angle_step*1.1547));	// the 1.1547 makes the overall distribution more like a hexagonal mesh
			h=(int)floor(h*sin(alt)+.5);
			if (h==0) h=1;
			h=fabs((float)csym)*floor(h/fabs((float)csym)+.5);
			h=2.0*PI/h;
			if (alt>0 && (az1-az0)/h<2.8) h=(az1-az0)/2.1;
			if (alt==0) h=az1;
			for (az=az0+(j%1)?h/2.0:0; az<az1-h/4; az+=h) {
				if (! with_mirror && az>PI && alt>PI/2-.001 && alt<PI/2.0+.001) continue;	// ignore half of the equtor
				float az_final = az;
				if(csym==-5) {	// for icos symmetry
					if(with_mirror) az_final+=3.0*PI/2.0-2*PI/10.0;
					else az_final+=3.0*PI/2.0;
				}
				Euler *e = new Euler(alt, az_final, 0);
				ret.push_back(e);
				//printf("%d\talt,az=%.2f, %.2f\n", ++i, alt*180./PI, az_final*180./PI);
			}
		}
	}
	return ret;

}


//                     INLINE FUNCTIONS
//inline float Euler::operator[](int b) { return angle[b];}

/* *****************************************************************************  */
// %% we need as many ways as possible to READ THE DATA %%

// set Euler from angles in MRC convention
void Euler::fromMRC(float theta, float phi, float omega){
	alt_=theta;	az_=fmod(-phi+5*PI/2,2*PI);	phi_=fmod(omega+PI/2,2*PI);
}

// return angles in MRC convention
float Euler::thetaMRC() const {return alt_;}
float Euler::phiMRC() const {return fmod(-az_ + 5*PI/2,2*PI);}
float Euler::omegaMRC() const {return fmod(phi_ +3* PI/2,2*PI);}

//        Internal and EMAN formulation of the rotation
float  Euler::alt() const { return alt_ ;}
float  Euler::az() const {return az_  ;}
float  Euler::phi() const { return phi_ ;}

float  Euler::alpha() const { return ((phi_) );}    // IMAGIC format Angles
float  Euler::beta() const { return alt_ ; }
float  Euler::gamma() const { return ((az_ ));}

float  Euler::phiSpider() const { 
#ifdef DEBUG0
	printf("phi=%f",(az_ + PI/2.+PI)); 
#endif
	return fmod((az_ + PI/2.+PI),PI*2.);
}
float  Euler::thetaSpider() const { 
#ifdef DEBUG0
	printf("theta=%f",(alt_)); 
#endif
	return alt_ ;
}
float  Euler::gammaSpider() const { 
#ifdef DEBUG0
	printf("gamma=%f",(phi_ - PI/2.+PI)); 
#endif
	return fmod((phi_ - PI/2.+PI),PI*2. );
}    // SPIDER format Angles

float  Euler::e0() const { return (  cos((phi_ + az_)/2.) * cos(alt_/2.)  );}  // Quaternion format
float  Euler::e1() const { return (  cos((phi_ - az_)/2.) * sin(alt_/2.)  );}
float  Euler::e2() const { return ( -sin((phi_ - az_)/2.) * sin(alt_/2.)  );}
float  Euler::e3() const { return ( sin((phi_ + az_)/2.) * cos(alt_/2.)  );}

// Spin-axis format
float  Euler::Q() const { return ( 2.* acos( cos((phi_ + az_)/2.) * cos(alt_/2.))  );}
float  Euler::n1() const { 
	float e0 =cos((phi_ + az_)/2.) * cos(alt_/2.)  ;
	if ((1-e0*e0)<1e-6) 
		return 0;
	else 
		return (  cos((phi_ - az_)/2.) * sin(alt_/2.) / (sqrt(1-e0*e0)) );
}
float  Euler::n2() const { 
	float e0 = cos((phi_ + az_)/2.) * cos(alt_/2.)  ;
	if ((1-e0*e0)<1e-6)
		return 0;
	else
		return ( -sin((phi_ - az_)/2.) * sin(alt_/2.) / (sqrt(1-e0*e0))  );
}
float  Euler::n3() const { 
	float e0 = cos((phi_ + az_)/2.) * cos(alt_/2.)  ;
	if ((1-e0*e0)<1e-6)
		return 1;
	else
		//return ( -sin((phi_ + az_)/2.) * cos(alt_/2.)/ (sqrt(1-e0*e0))   );
		return ( sin((phi_ + az_)/2.) * cos(alt_/2.)/ (sqrt(1-e0*e0))   );
}


// SGI format
float  Euler::Qsgi() const { return ( 2.* acos( cos((phi_ + az_)/2.) * cos(alt_/2.))  );}
float  Euler::n1sgi() const { float e0 =cos((phi_ + az_)/2.) * cos(alt_/2.)  ;
					return (  -cos((phi_ - az_)/2.) * sin(alt_/2.) / (sqrt(1-e0*e0)) );}
float  Euler::n2sgi() const { float e0 = cos((phi_ + az_)/2.) * cos(alt_/2.)  ;
					return ( -sin((phi_ - az_)/2.) * sin(alt_/2.) / (sqrt(1-e0*e0))  );}
float  Euler::n3sgi() const { float e0 = cos((phi_ + az_)/2.) * cos(alt_/2.)  ;
					return ( sin((phi_ + az_)/2.) * cos(alt_/2.)/ (sqrt(1-e0*e0))   );}

// Matrix format for a rotation
float *Euler::mat() { 
	m_[0]=  cos(phi_) * cos(az_) - sin(phi_) * cos(alt_)*sin(az_);
	m_[1]=  cos(phi_) * sin(az_) + sin(phi_) * cos(alt_)*cos(az_);
	m_[2]=  sin(phi_) * sin(alt_);
	m_[3]= -sin(phi_) * cos(az_) - cos(phi_) * cos(alt_)*sin(az_);
	m_[4]= -sin(phi_) * sin(az_) + cos(phi_) * cos(alt_)*cos(az_);
	m_[5]=  cos(phi_) * sin(alt_);
	m_[6]=  sin(alt_) * sin(az_) ;
	m_[7]= -sin(alt_) * cos(az_) ;
	m_[8]=  cos(alt_);
	return m_;
}		




float * matmulmat(float *mleft, float *mright) {
	float *mreturn;
	int i,j,k;	
	mreturn = (float *) malloc(9*sizeof(float));
	for (i=0; i <3; i++){ 
		for (j=0; j<3; j++){ 
			mreturn[j+3*i]=0;
			for (k=0;k<3;k++){
				mreturn[j+3*i] += mleft[k+3*i] * mright[j+3*k];
	}}} // end k j i loops	
	return mreturn;
} // end matmulmat

void Euler::print(char *id, FILE* out){
fprintf(out,"Euler %s =( %.4f , %.4f , %.4f )\n",id,alt_*180/PI,az_*180/PI,phi_*180/PI);
}	
	
void Euler::printMatrix(char *id, FILE* out){
mat();
fprintf(out,"Euler %s\n",id);
for(int i=0;i<3;i++){
	for(int j=0;j<3;j++){
		fprintf(out,"%f\t",m_[i*3+j]);
	}
	fprintf(out,"\n");
}
fprintf(out,"\n");
}	

///////////////*******************************//////////////////////////
//////// the following 5 functions are added by Xiangan Liu (02/20/2004)
///////////////*******************************//////////////////////////

Vect GetXCoef();      //the first  line of the transform matrix
Vect GetYCoef();      //the second line of the transform matrix
Vect GetZCoef();      //the third  line of the transform matrix
Vect GetNormal();  // get Z' axis coordinates in XYZ system
Vect TransformXYZtoXYZprime ( const Vect &v ); //transform a vector from XYZ system to X'Y'Z' system

// Order of rotation for EMAN Euler angle is: z (by az), x' (by alt), z'' (by phi)
Vect Euler::GetXCoef() {  //the first line of the transform matrix
   return Vect ( cos(az_)  * cos(phi_) - cos(alt_) * sin(az_) * sin(phi_),
                 sin(az_)  * cos(phi_) + cos(alt_) * cos(az_) * sin(phi_),
                 sin(alt_) * sin(phi_) );
}

Vect Euler::GetYCoef() {  //the second line of the transform matrix
   return Vect ( -cos(az_)  * sin(phi_) - cos(alt_) * sin(az_) * cos(phi_),
                 -sin(az_)  * sin(phi_) + cos(alt_) * cos(az_) * cos(phi_) ,
                  sin(alt_) * cos(phi_) );
}

Vect Euler::GetZCoef() {  //the third line of the transform matrix
   return Vect (  sin(alt_) * sin(az_),
                 -sin(alt_) * cos(az_),
                  cos(alt_) );
}

Vect Euler::GetNormal () {
   return Vect (  sin(alt_) * sin(az_),
                 -sin(alt_) * cos(az_),
                  cos(alt_) );
}


Vect Euler::TransformXYZtoXYZprime ( Vect &v ) {
   return Vect(GetXCoef().inner(v),
               GetYCoef().inner(v),
               GetZCoef().inner(v));

}


//	MRC euler angle to EMAN euler angle
//	Meman = Mmrc Meman2mrc
//	Meman(alt_eman, az_eman, phi_eman) = Meman(alt_mrc, -az_mrc+90, phi_mrc+180) Meman(31.7174, -90,  90)
//							alt_eman =  alt_mrc
//							az_eman  = -az_mrc  + 90
//							phi_eman =  phi_mrc + 180
void Euler::setAngleFromMRC(float alt_mrc, float az_mrc, float phi_mrc){
	//Euler EUmrc(alt_mrc, -az_mrc + 1.570796326794897, phi_mrc + 3.1415926535897);
	//Euler EUeman2mrc(0.553574358897, -1.570796326794897, 1.570796326794897); //31.7174 degree = 0.553574358897 rad

	//Euler EUmrc(alt_mrc, -az_mrc + 1.570796326794897, phi_mrc + 1.570796326794897);
	//Euler EUeman2mrc(1.0172219678978513677, -1.570796326794897, 0); //31.7174 degree = 0.553574358897 rad

	Euler EUmrc(alt_mrc, az_mrc, phi_mrc);
	Euler EUeman2mrc(0.553574358897, 1.570796326794897, -1.570796326794897); //31.7174 degree = 0.553574358897 rad

	float *Matrix = matmulmat(EUmrc.mat(), EUeman2mrc.mat());
	setAngle(Matrix);
	free(Matrix);
}


//	EMAN euler angle to MRC euler angle
//	Mmrc = Meman / Meman2mrc = Meman Mmrc2eman
//	Meman(alt_eman, -az_eman+90, phi_eman+180) = Meman(alt_eman, az_eman, phi_eman) Meman(31.7174,  90, -90)
//		alt_mrc =  alt_eman
//		az_mrc  = -az_eman  + 90
//		phi_mrc =  phi_eman + 180
void Euler::convertToMRCAngle(){
/*
	Euler EUmrc2eman(0.553574358897, -1.570796326794897, 1.570796326794897); //31.7174 degree = 0.553574358897 rad
	float *Matrix = matmulmat(mat(), EUmrc2eman.mat());
	Euler EUmrc(Matrix);
	alt_mrc =        EUmrc.alt();
	az_mrc  =  fmod(-EUmrc.az()  + 1.570796326794897*5, 6.2831853071796);
	phi_mrc =  fmod( EUmrc.phi() + 3.1415926535897, 6.2831853071796);
*/
	Euler EUmrc2eman(0.553574358897, -1.570796326794897, 1.570796326794897); //31.7174 degree = 0.553574358897 rad
	float *Matrix = matmulmat(mat(), EUmrc2eman.mat());
	Euler EUmrc(Matrix);
	alt_mrc =        EUmrc.alt();
	az_mrc  =  fmod(-EUmrc.az()  + 1.570796326794897*5, 6.2831853071796);
	phi_mrc =  fmod( EUmrc.phi() + 1.570796326794897*3, 6.2831853071796);
	
	if (alt_mrc > PI/2.0) { // use the equivalent view that is in the same asymmetric unit used by MRC programs
		alt_mrc = PI - alt_mrc;
		az_mrc  = PI - az_mrc;
		phi_mrc = phi_mrc - PI;
	}
	if (az_mrc  > PI)   az_mrc  -= 2*PI;
	if (phi_mrc > PI)   phi_mrc -= 2*PI;

	free(Matrix);
}

// return euler angles of MRC converted from eman convention, before using them, need to call convetTOMRCAngle()
float Euler::alt_MRC() {return alt_mrc;}
float Euler::az_MRC()  {return az_mrc;}
float Euler::phi_MRC() {return phi_mrc;}

